import { Color } from "./engine/datatypes.ts";
import { Stat } from "./datatypes.ts";
import { maybeGetHuntMode } from "./huntmode.ts";
import { getEndgameModal } from "./endgamemodal.ts";
import { getVNModal } from "./vnmodal.ts";

export type Microtheme = {
  SKY0: Color; // outer, less dark
  FLOOR0: Color; // floor, even less dark

  WALL0: Color; // darkest (ex. the underside of something)
  WALL1: Color; // darkest (ex. the underside of something)

  BG0: Color; // UI background -- should be highly readable and similar to SKY or FLOOR
  FG1: Color; // dark (ex. disabled text)
  FG2: Color; // normal  (ex. normal text)
  FG3: Color; // brightest (ex. bold text)
};

/*
const MICROTHEME_OLD: MicroTheme = {
  BG0: Color.parseHexCode("#242234"),
  BG1: Color.parseHexCode("#143464"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
}
 */

export const MICROTHEME_DEFAULT: Microtheme = {
  // outdoors
  SKY0: Color.parseHexCode("#242234"),
  FLOOR0: Color.parseHexCode("#141013"),
  WALL0: Color.parseHexCode("#5d758d"),
  WALL1: Color.parseHexCode("#8b93af"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_BLACK: Microtheme = {
  // manor
  SKY0: Color.parseHexCode("#141013"),
  FLOOR0: Color.parseHexCode("#3b1725"),
  WALL0: Color.parseHexCode("#221c1a"),
  WALL1: Color.parseHexCode("#322b28"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_PURPLE_TAN: Microtheme = {
  // library
  SKY0: Color.parseHexCode("#403353"),
  FLOOR0: Color.parseHexCode("#242234"),
  WALL0: Color.parseHexCode("#5a4e44"),
  WALL1: Color.parseHexCode("#c7b08b"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_GREEN: Microtheme = {
  // zoo
  SKY0: Color.parseHexCode("#24523b"),
  FLOOR0: Color.parseHexCode("#122020"),
  WALL0: Color.parseHexCode("#328464"),
  WALL1: Color.parseHexCode("#5daf8d"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_TEAL: Microtheme = {
  // optometrist
  SKY0: Color.parseHexCode("#143464"),
  FLOOR0: Color.parseHexCode("#122020"),
  WALL0: Color.parseHexCode("#477d85"),
  WALL1: Color.parseHexCode("#588dbe"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_PINK: Microtheme = {
  // coffee shop
  SKY0: Color.parseHexCode("#793a80"),
  FLOOR0: Color.parseHexCode("#221c1a"),
  WALL0: Color.parseHexCode("#e86a73"),
  WALL1: Color.parseHexCode("#f5a097"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_NEON: Microtheme = {
  // club
  SKY0: Color.parseHexCode("#141013"),
  FLOOR0: Color.parseHexCode("#221c1a"),
  WALL0: Color.parseHexCode("#9cdb43"),
  WALL1: Color.parseHexCode("#d6f264"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export const MICROTHEME_RED: Microtheme = {
  // blood bank
  SKY0: Color.parseHexCode("#73172d"),
  FLOOR0: Color.parseHexCode("#141013"),
  WALL0: Color.parseHexCode("#df3e23"),
  WALL1: Color.parseHexCode("#f9a31b"),
  BG0: Color.parseHexCode("#000000"),
  FG1: Color.parseHexCode("#8b93af"),
  FG2: Color.parseHexCode("#b3b9d1"),
  FG3: Color.parseHexCode("#ffffff"),
};

export class ColorSystem {
  get BG_UI() {
    return this.#microtheme.BG0;
  }
  get BG_FLOOR() {
    return this.#microtheme.FLOOR0;
  }
  get FG_TEXT() {
    return this.#microtheme.FG2;
  }
  get FG_TEXT_DISABLED() {
    return this.#microtheme.FG1;
  }
  readonly FG_TOO_EXPENSIVE = Color.parseHexCode("#f5a097");
  readonly FG_TEXT_ENDORSED = Color.parseHexCode("#d6f264");
  get FG_BOLD() {
    return this.#microtheme.FG3;
  }
  get BG_OUTER() {
    return this.#microtheme.SKY0;
  }
  get BG_WALL_OR_UNREVEALED() {
    return this.#microtheme.SKY0;
  }
  get BG_CEILING() {
    return this.#microtheme.SKY0;
  }

  get BG_INNERWALL() {
    return this.#microtheme.WALL0;
  }
  get BG_OUTERWALL() {
    return this.#microtheme.WALL1;
  }

  get #microtheme(): Microtheme {
    if (getEndgameModal().isShown || getVNModal().isShown) {
      return MICROTHEME_RED;
    }
    let option = maybeGetHuntMode()?.getActiveMicrotheme();
    if (option) {
      return option;
    }
    return MICROTHEME_DEFAULT;
  }

  // stat colors
  readonly SWATCH_EXP: [Color, Color] = [
    Color.parseHexCode("#b9bffb"),
    Color.parseHexCode("#e3e6ff"),
  ];

  readonly SWATCH_AGI: [Color, Color] = [
    Color.parseHexCode("#df3e23"),
    Color.parseHexCode("#fa6a0a"),
  ];

  readonly SWATCH_INT: [Color, Color] = [
    Color.parseHexCode("#285cc4"),
    Color.parseHexCode("#249fde"),
  ];

  readonly SWATCH_CHA: [Color, Color] = [
    Color.parseHexCode("#793a80"),
    Color.parseHexCode("#bc4a9b"),
  ];

  readonly SWATCH_PSI: [Color, Color] = [
    Color.parseHexCode("#9cdb43"),
    Color.parseHexCode("#d6f264"),
  ];

  readonly SWATCH_STAT: Record<Stat, [Color, Color]> = {
    AGI: this.SWATCH_AGI,
    INT: this.SWATCH_INT,
    CHA: this.SWATCH_CHA,
    PSI: this.SWATCH_PSI,
  };
}

export let C = new ColorSystem();
